@@ -10,7 +10,7 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	22 November 1989
- * Last Edited:	31 January 1996
+ * Last Edited:	3 June 1996
  *
  * Copyright 1996 by the University of Washington
  *
@@ -662,7 +662,7 @@ long mail_delete (stream,mailbox)
 	char *mailbox;
 {
   DRIVER *factory = mail_valid (stream,mailbox,"delete mailbox");
-  return factory ? (*factory->delete) (stream,mailbox) : NIL;
+  return factory ? (*factory->mbxdel) (stream,mailbox) : NIL;
 }
 
 
@@ -685,7 +685,7 @@ long mail_rename (stream,old,new)
     mm_log (tmp,ERROR);
     return NIL;
   }
-  return factory ? (*factory->rename) (stream,old,new) : NIL;
+  return factory ? (*factory->mbxren) (stream,old,new) : NIL;
 }
 
 /* Mail open
