@@ -1,5 +1,5 @@
 #if !defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: pine.c,v 4.330 1996/04/09 22:42:18 mikes Exp $";
+static char rcsid[] = "$Id: pine.c,v 4.334 1996/05/16 23:43:36 hubert Exp $";
 #endif
 /*----------------------------------------------------------------------
 
@@ -1097,8 +1097,8 @@ main_menu_screen(pine_state)
 #endif
         orig_ch = ch;
 
-        if(ch < 0x0100 && isupper(ch))
-          ch = tolower(ch);
+        if(ch < 0x0100 && isupper((unsigned char)ch))
+          ch = tolower((unsigned char)ch);
 	else if(ch >= PF1 && ch <= PF12 && km->which == 1)
 	  ch = PF2OPF(ch);
 
@@ -1107,8 +1107,8 @@ main_menu_screen(pine_state)
 	  ch = F_ON(F_USE_FK,pine_state)
                                        ? mkeys[current_default_menu_item].f_key
                                        : mkeys[current_default_menu_item].key;
-          if(ch <= 0xff && isupper(ch))
-            ch = tolower(ch);
+          if(ch <= 0xff && isupper((unsigned char)ch))
+            ch = tolower((unsigned char)ch);
 	}
 
 
@@ -2030,7 +2030,7 @@ phone_home_blurb(dline)
 {
     static char *blurb[] = {
       "SPECIAL OFFER:  Would you like to receive (via email) a brief document",
-      "  entitled \"Getting the most out of Pine 3.92\" ?"
+      "  entitled \"Getting the most out of Pine 3.94\" ?"
     };
 
     if(dline+3 < ps_global->ttyo->screen_rows - FOOTER_ROWS(ps_global))
@@ -2135,9 +2135,6 @@ goodnight_gracey(pine_state, exit_val)
 
     cur_is_inbox = (pine_state->inbox_stream == pine_state->mail_stream);
 
-    /* committed to exiting so become unreceptive to HUP */
-    end_sighup();
-
     /* clean up open streams */
     if(pine_state->mail_stream)
       expunge_and_close(pine_state->mail_stream, pine_state->context_current,
